from evmosgrpc.messages.msgsend import create_msg_send


def test_create_msg_send():
    a = create_msg_send('evmos1h8057cmr2znknsaqu6vqazfughcfq93k24dmw4', 'evmos1dlt6lgs6sslurfl3zyl7tgmnvzhq5us2uquxv7',
                        100, 'evmos')
    assert a.SerializeToString(
    ) == b'\n,evmos1h8057cmr2znknsaqu6vqazfughcfq93k24dmw4\x12,evmos1dlt6lgs6sslurfl3zyl7tgmnvzhq5us2uquxv7\x1a\x0c\n\x05evmos\x12\x03100'  # NOQA E501



import os

GRPC_ENDPOINT = os.getenv('GRPC_ENDPOINT', '127.0.0.1:9090')
MEMO = os.getenv('MEMO', 'Sent from Hanchon\'s faucet')
FEE = os.getenv('FEE', '20')
GAS_LIMIT = os.getenv('GAS_LIMIT', '200000')
CHAIN_ID = os.getenv('CHAIN_ID', 'evmos_9000-1')
DENOM = os.getenv('DENOM', 'aphoton')

# Types
ETHSECP256K1 = 'ethsecp256k1'
SECP256K1 = 'secp256k1'



import grpc
from evmoswallet import Wallet
from google.protobuf.message import Message

from evmosgrpc.accounts import get_account_grpc
from evmosgrpc.broadcaster import broadcast
from evmosgrpc.constants import ETHSECP256K1
from evmosgrpc.constants import SECP256K1


class TransactionBuilder():
    def __init__(self, seed: str) -> None:
        self.wallet = Wallet(seed)
        self.algo = ETHSECP256K1
        self.address = self.wallet.evmos_address
        try:
            self.account_number, self.sequence, _ = get_account_grpc(self.address)
        except grpc._channel._InactiveRpcError:
            self.account_number = self.sequence = None

    def send_tx(self, msg: Message):
        self.update_sequence()
        return broadcast(msg)

    def update_sequence(self):
        _, self.sequence, _ = get_account_grpc(self.address)


class PubkeyWallet():
    def __init__(self, pubkey) -> None:
        self.public_key = pubkey


class ExternalWallet(TransactionBuilder):
    def __init__(self, address: str, algo: str = SECP256K1, pubkey=None) -> None:
        self.account_number, self.sequence, pubkey_grpc = get_account_grpc(self.address)
        if pubkey is not None:
            self.wallet = PubkeyWallet(pubkey_grpc)
        else:
            self.wallet = PubkeyWallet(pubkey)
        self.algo = algo
        self.address = address



from evmosproto.cosmos.base.v1beta1.coin_pb2 import Coin
from evmosproto.cosmos.crypto.secp256k1.keys_pb2 import PubKey as Secp256PubKey
from evmosproto.cosmos.tx.signing.v1beta1.signing_pb2 import SIGN_MODE_DIRECT
from evmosproto.cosmos.tx.v1beta1.tx_pb2 import AuthInfo
from evmosproto.cosmos.tx.v1beta1.tx_pb2 import Fee
from evmosproto.cosmos.tx.v1beta1.tx_pb2 import ModeInfo
from evmosproto.cosmos.tx.v1beta1.tx_pb2 import SignDoc
from evmosproto.cosmos.tx.v1beta1.tx_pb2 import SignerInfo
from evmosproto.cosmos.tx.v1beta1.tx_pb2 import TxBody
from evmosproto.cosmos.tx.v1beta1.tx_pb2 import TxRaw
from evmosproto.ethermint.crypto.v1.ethsecp256k1.keys_pb2 import PubKey
from evmosproto.google.protobuf.any_pb2 import Any
from evmoswallet.eth.ethereum import sha3_256
from google.protobuf.message import Message

from evmosgrpc.builder import TransactionBuilder
from evmosgrpc.constants import CHAIN_ID
from evmosgrpc.constants import DENOM
from evmosgrpc.constants import ETHSECP256K1
from evmosgrpc.constants import FEE
from evmosgrpc.constants import GAS_LIMIT
from evmosgrpc.constants import MEMO
from evmosgrpc.constants import SECP256K1


class Transaction:
    def create_body_bytes(self, msg: Message):
        body = TxBody()
        any = Any()
        any.Pack(msg, type_url_prefix='/')
        body.messages.append(any)
        body.memo = MEMO
        self.body = body

    def create_fee(self):
        coin = Coin()
        coin.denom = DENOM
        coin.amount = FEE
        fee = Fee()
        fee.amount.append(coin)
        fee.gas_limit = int(GAS_LIMIT)
        self.fee = fee

    def create_signer_info(self):
        signer_info = SignerInfo()
        if self.builder.algo == ETHSECP256K1:
            pub_key = PubKey()
        elif self.builder.algo == SECP256K1:
            pub_key = Secp256PubKey()
        else:
            raise NotImplementedError(f'{self.builer.algo} not supported.')

        pub_key.key = self.builder.wallet.public_key
        public_key = Any()

        public_key.Pack(pub_key, type_url_prefix='/')
        signer_info.public_key.CopyFrom(public_key)

        a = ModeInfo()
        single = ModeInfo.Single()
        single.mode = SIGN_MODE_DIRECT
        a.single.CopyFrom(single)
        signer_info.mode_info.CopyFrom(a)
        signer_info.sequence = int(self.builder.sequence)

        self.signer_info = signer_info

    def create_auth_info_bytes(self):
        auth_info = AuthInfo()
        auth_info.signer_infos.append(self.signer_info)
        auth_info.fee.CopyFrom(self.fee)
        self.info = auth_info

    def create_sig_doc(self):
        doc = SignDoc()
        doc.body_bytes = self.body.SerializeToString()
        doc.auth_info_bytes = self.info.SerializeToString()
        doc.chain_id = CHAIN_ID
        doc.account_number = int(self.builder.account_number)
        return doc.SerializeToString()

    def create_signatures(self):
        to_sign = self.create_sig_doc()
        return self.builder.wallet.sign(sha3_256(to_sign).digest())

    def create_tx_raw_with_class_info(self, signature):
        return create_tx_raw(self.body.SerializeToString(), self.info.SerializeToString(), signature)

    def create_tx_template(self, builder: TransactionBuilder, msg: Message):
        self.builder = builder
        self.create_body_bytes(msg)
        self.create_fee()
        self.create_signer_info()
        self.create_auth_info_bytes()

    def generate_tx(self, builder: TransactionBuilder, msg: Message):
        self.create_tx_template(builder, msg)
        return self.create_tx_raw_with_class_info(self.create_signatures())


def create_tx_raw(body_bytes, auth_info, signature):
    tx = TxRaw()
    tx.body_bytes = body_bytes
    tx.auth_info_bytes = auth_info
    tx.signatures.append(signature)
    return tx



from evmosproto.cosmos.auth.v1beta1.query_pb2 import QueryAccountRequest
from evmosproto.cosmos.auth.v1beta1.query_pb2_grpc import QueryStub
from evmosproto.ethermint.types.v1.account_pb2 import EthAccount
from google.protobuf.json_format import MessageToDict

from evmosgrpc.utils import create_grpc_channel


# Get account, returns pub_key, account_number, sequence
def get_account_grpc(account: str):
    # Create channel
    stub = QueryStub(create_grpc_channel())
    params = QueryAccountRequest()
    params.address = account
    # Send the request
    resp = stub.Account(params)
    # Parse the Any message
    account = EthAccount()
    account.ParseFromString(resp.account.value)
    account = MessageToDict(account, including_default_value_fields=True)['baseAccount']

    pubkey = None
    if 'pubKey' in account:
        pubkey = account['pubKey']['key']

    return account['accountNumber'], account['sequence'], pubkey



import json

from evmosproto.cosmos.bank.v1beta1.tx_pb2 import MsgSend
from google.protobuf.json_format import Parse

from evmosgrpc.constants import DENOM


def create_msg_send(origin, dest, amount, denom=DENOM):
    raw_msg = {
        'fromAddress': origin,
        'toAddress': dest,
        'amount': [{
            'denom': denom,
            'amount': str(amount),
        }],
    }
    return Parse(json.dumps(raw_msg), MsgSend())



from evmosproto.cosmos.tx.v1beta1.service_pb2 import BROADCAST_MODE_SYNC
from evmosproto.cosmos.tx.v1beta1.service_pb2 import BroadcastTxRequest
from evmosproto.cosmos.tx.v1beta1.service_pb2_grpc import ServiceStub
from google.protobuf.message import Message

from evmosgrpc.utils import create_grpc_channel


def create_broadcast_tx(tx: Message, mode=BROADCAST_MODE_SYNC):
    broadcast_tx = BroadcastTxRequest()
    broadcast_tx.tx_bytes = tx.SerializeToString()
    broadcast_tx.mode = BROADCAST_MODE_SYNC
    return broadcast_tx


def broadcast(tx: Message, mode=BROADCAST_MODE_SYNC):
    tx = create_broadcast_tx(tx, mode)
    channel = create_grpc_channel()
    stub = ServiceStub(channel)
    return stub.BroadcastTx(tx)
